"""New traits for our custom serialization.

Refer to https://traitlets.readthedocs.io/en/stable/defining_traits.html for
documentation on how to define new traitlet types.
"""

from __future__ import annotations

from typing import Any, List, Set, Tuple, Union

import matplotlib as mpl
import numpy as np
import pyarrow as pa
import traitlets
from traitlets import TraitError
from traitlets.traitlets import TraitType
from traitlets.utils.descriptions import class_of, describe
from typing_extensions import Self

from lonboard._serialization import (
    ACCESSOR_SERIALIZATION,
    TABLE_SERIALIZATION,
)
from lonboard._utils import get_geometry_column_index


# This is a custom subclass of traitlets.TraitType because its `error` method ignores
# the `info` passed in. See https://github.com/developmentseed/lonboard/issues/71 and
# https://github.com/ipython/traitlets/pull/884
class FixedErrorTraitType(traitlets.TraitType):
    def error(self, obj: Self, value, error=None, info=None):
        """Raise a TraitError

        Parameters
        ----------
        obj : HasTraits or None
            The instance which owns the trait. If not
            object is given, then an object agnostic
            error will be raised.
        value : any
            The value that caused the error.
        error : Exception (default: None)
            An error that was raised by a child trait.
            The arguments of this exception should be
            of the form ``(value, info, *traits)``.
            Where the ``value`` and ``info`` are the
            problem value, and string describing the
            expected value. The ``traits`` are a series
            of :class:`TraitType` instances that are
            "children" of this one (the first being
            the deepest).
        info : str (default: None)
            A description of the expected value. By
            default this is infered from this trait's
            ``info`` method.
        """
        if error is not None:
            # handle nested error
            error.args += (self,)
            if self.name is not None:
                # this is the root trait that must format the final message
                chain = " of ".join(describe("a", t) for t in error.args[2:])
                if obj is not None:
                    error.args = (
                        "The '{}' trait of {} instance contains {} which "
                        "expected {}, not {}.".format(
                            self.name,
                            describe("an", obj),
                            chain,
                            error.args[1],
                            describe("the", error.args[0]),
                        ),
                    )
                else:
                    error.args = (
                        "The '{}' trait contains {} which "
                        "expected {}, not {}.".format(
                            self.name,
                            chain,
                            error.args[1],
                            describe("the", error.args[0]),
                        ),
                    )
            raise error
        else:
            # this trait caused an error
            if self.name is None:
                # this is not the root trait
                raise TraitError(value, info or self.info(), self)
            else:
                # this is the root trait
                if obj is not None:
                    e = "The '{}' trait of {} instance expected {}, not {}.".format(
                        self.name,
                        class_of(obj),
                        # CHANGED:
                        # Use info if provided
                        info or self.info(),
                        describe("the", value),
                    )
                else:
                    e = "The '{}' trait expected {}, not {}.".format(
                        self.name,
                        # CHANGED:
                        # Use info if provided
                        info or self.info(),
                        describe("the", value),
                    )
                raise TraitError(e)


class PyarrowTableTrait(FixedErrorTraitType):
    """A trait to validate input for a geospatial Arrow-backed table

    Allowed input includes:

    - A pyarrow [`Table`][pyarrow.Table] or containing a geometry column with [GeoArrow metadata](https://geoarrow.org/extension-types).
    - Any GeoArrow table from a library that implements the [Arrow PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
      This includes the
      [`GeoTable`](https://geoarrow.github.io/geoarrow-rs/python/latest/api/core/table/#geoarrow.rust.core.GeoTable)
      class from
      [`geoarrow-rust`](https://geoarrow.github.io/geoarrow-rs/python/latest/).
    """

    default_value = None
    info_text = "a pyarrow or GeoArrow Table"

    def __init__(
        self: TraitType,
        *args,
        allowed_geometry_types: Set[bytes] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(
            sync=True,
            allowed_geometry_types=allowed_geometry_types,
            **TABLE_SERIALIZATION,
        )

    def validate(self, obj: Self, value: Any):
        if not isinstance(value, pa.Table):
            self.error(obj, value)

        allowed_geometry_types = self.metadata.get("allowed_geometry_types")
        # No restriction on the allowed geometry types in this table
        if not allowed_geometry_types:
            return value

        geom_col_idx = get_geometry_column_index(value.schema)
        geometry_extension_type = value.schema.field(geom_col_idx).metadata.get(
            b"ARROW:extension:name"
        )

        if (
            allowed_geometry_types
            and geometry_extension_type not in allowed_geometry_types
        ):
            allowed_types_str = "\n".join(allowed_geometry_types)
            msg = (
                f"Expected one of {allowed_types_str} geometry types, "
                "got {geometry_extension_type}."
            )
            self.error(obj, value, info=msg)

        return value


class ColorAccessor(FixedErrorTraitType):
    """A trait to validate input for a deck.gl color accessor.

    Various input is allowed:

    - A `list` or `tuple` with three or four integers, ranging between 0 and 255
      (inclusive). This will be used as the color for all objects.
    - A `str` representing a hex color or "well known" color interpretable by
      [matplotlib.colors.to_rgba][matplotlib.colors.to_rgba].
    - A numpy `ndarray` with two dimensions and data type [`np.uint8`][numpy.uint8]. The
      size of the second dimension must be `3` or `4`, and will correspond to either RGB
      or RGBA colors.
    - A pyarrow [`FixedSizeListArray`][pyarrow.FixedSizeListArray] or
      [`ChunkedArray`][pyarrow.ChunkedArray] containing `FixedSizeListArray`s. The inner
      size of the fixed size list must be `3` or `4` and its child must have type
      `uint8`.
    - Any Arrow fixed size list array from a library that implements the [Arrow
      PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).

    You can use helpers in the `lonboard.colormap` module (i.e.
    [`apply_continuous_cmap`][lonboard.colormap.apply_continuous_cmap]) to simplify
    constructing numpy arrays for color values.
    """

    default_value = (0, 0, 0)
    info_text = (
        "a tuple or list representing an RGB(A) color or numpy ndarray or "
        "pyarrow FixedSizeList representing an array of RGB(A) colors"
    )

    def __init__(
        self: TraitType,
        *args,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def validate(
        self, obj, value
    ) -> Union[Tuple[int, ...], List[int], pa.ChunkedArray, pa.FixedSizeListArray]:
        if isinstance(value, (tuple, list)):
            if len(value) < 3 or len(value) > 4:
                self.error(obj, value, info="3 or 4 values if passed a tuple or list")

            if any(not isinstance(v, int) for v in value):
                self.error(
                    obj,
                    value,
                    info="all values to be integers if passed a tuple or list",
                )

            if any(v < 0 or v > 255 for v in value):
                self.error(
                    obj,
                    value,
                    info="values between 0 and 255",
                )

            return value

        if isinstance(value, np.ndarray):
            if not np.issubdtype(value.dtype, np.uint8):
                self.error(obj, value, info="Color array must be uint8 type.")

            if value.ndim != 2:
                self.error(obj, value, info="Color array must have 2 dimensions.")

            list_size = value.shape[1]
            if list_size not in (3, 4):
                self.error(
                    obj,
                    value,
                    info="Color array must have 3 or 4 as its second dimension.",
                )

            return pa.FixedSizeListArray.from_arrays(value.flatten("C"), list_size)

        # Check for Arrow PyCapsule Interface
        # https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html
        # TODO: with pyarrow v16 also import chunked array from stream
        if not isinstance(value, (pa.ChunkedArray, pa.Array)):
            if hasattr(value, "__arrow_c_array__"):
                value = pa.array(value)

        if isinstance(value, (pa.ChunkedArray, pa.Array)):
            if not pa.types.is_fixed_size_list(value.type):
                self.error(
                    obj, value, info="Color pyarrow array must be a FixedSizeList."
                )

            if value.type.list_size not in (3, 4):
                self.error(
                    obj,
                    value,
                    info=(
                        "Color pyarrow array must have a FixedSizeList inner size of "
                        "3 or 4."
                    ),
                )

            if not pa.types.is_uint8(value.type.value_type):
                self.error(
                    obj, value, info="Color pyarrow array must have a uint8 child."
                )

            return value

        if isinstance(value, str):
            try:
                c = mpl.colors.to_rgba(value)  # type: ignore
            except ValueError:
                self.error(
                    obj,
                    value,
                    info=(
                        "Color string must be a hex string interpretable by "
                        "matplotlib.colors.to_rgba."
                    ),
                )
                return

            return tuple(map(int, (np.array(c) * 255).astype(np.uint8)))

        self.error(obj, value)
        assert False


class FloatAccessor(FixedErrorTraitType):
    """A trait to validate input for a deck.gl float accessor.

    Various input is allowed:

    - An `int` or `float`. This will be used as the value for all objects.
    - A numpy `ndarray` with a numeric data type. This will be casted to an array of
      data type [`np.float32`][numpy.float32]. Each value in the array will be used as
      the value for the object at the same row index.
    - A pandas `Series` with a numeric data type. This will be casted to an array of
      data type [`np.float32`][numpy.float32]. Each value in the array will be used as
      the value for the object at the same row index.
    - A pyarrow [`FloatArray`][pyarrow.FloatArray], [`DoubleArray`][pyarrow.DoubleArray]
      or [`ChunkedArray`][pyarrow.ChunkedArray] containing either a `FloatArray` or
      `DoubleArray`. Each value in the array will be used as the value for the object at
      the same row index.
    - Any Arrow floating point array from a library that implements the [Arrow PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
      This includes data structures from
      [`geoarrow-rust`](https://geoarrow.github.io/geoarrow-rs/python/latest/).
    """

    default_value = float(0)
    info_text = (
        "a float value or numpy ndarray or pyarrow array representing an array"
        " of floats"
    )

    def __init__(
        self: TraitType,
        *args,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def validate(self, obj, value) -> Union[float, pa.ChunkedArray, pa.DoubleArray]:
        if isinstance(value, (int, float)):
            return float(value)

        # pandas Series
        if (
            value.__class__.__module__.startswith("pandas")
            and value.__class__.__name__ == "Series"
        ):
            # Cast pandas Series to numpy ndarray
            value = np.asarray(value)

        if isinstance(value, np.ndarray):
            if not np.issubdtype(value.dtype, np.number):
                self.error(obj, value, info="numeric dtype")

            # TODO: should we always be casting to float32? Should it be
            # possible/allowed to pass in ~int8 or a data type smaller than float32?
            return pa.array(value.astype(np.float32))

        # Check for Arrow PyCapsule Interface
        # https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html
        # TODO: with pyarrow v16 also import chunked array from stream
        if not isinstance(value, (pa.ChunkedArray, pa.Array)):
            if hasattr(value, "__arrow_c_array__"):
                value = pa.array(value)

        if isinstance(value, (pa.ChunkedArray, pa.Array)):
            if not pa.types.is_floating(value.type):
                self.error(
                    obj,
                    value,
                    info="Float pyarrow array must be a floating point type.",
                )

            return value.cast(pa.float32())

        self.error(obj, value)
        assert False


class TextAccessor(FixedErrorTraitType):
    """A trait to validate input for a deck.gl text accessor.

    Various input is allowed:

    - A `str`. This will be used as the value for all objects.
    - A numpy `ndarray` with a string data type Each value in the array will be used as
      the value for the object at the same row index.
    - A pandas `Series` with a string data type. Each value in the array will be used as
      the value for the object at the same row index.
    - A pyarrow [`StringArray`][pyarrow.StringArray] or
      [`ChunkedArray`][pyarrow.ChunkedArray] containing a `StringArray`. Each value in
      the array will be used as the value for the object at the same row index.
    """

    default_value = ""
    info_text = (
        "a string value or numpy ndarray or pandas Series or pyarrow array representing"
        " an array of strings"
    )

    def __init__(
        self: TraitType,
        *args,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def validate(self, obj, value) -> Union[float, pa.ChunkedArray, pa.DoubleArray]:
        if isinstance(value, str):
            return value

        # pandas Series
        if (
            value.__class__.__module__.startswith("pandas")
            and value.__class__.__name__ == "Series"
        ):
            # Cast pandas Series to pyarrow array
            value = pa.array(value)

        if isinstance(value, np.ndarray):
            value = pa.StringArray.from_pandas(value)

        if isinstance(value, (pa.ChunkedArray, pa.Array)):
            if not pa.types.is_string(value.type):
                self.error(
                    obj,
                    value,
                    info="String pyarrow array must be a string type.",
                )

            return value

        self.error(obj, value)
        assert False



from __future__ import annotations

from pathlib import Path
from typing import Sequence, Union

import ipywidgets
import traitlets
from ipywidgets.embed import embed_minimal_html

from lonboard._base import BaseAnyWidget
from lonboard._environment import DEFAULT_HEIGHT
from lonboard._layer import BaseLayer
from lonboard._viewport import compute_view
from lonboard.basemap import CartoBasemap

# bundler yields lonboard/static/{index.js,styles.css}
bundler_output_dir = Path(__file__).parent / "static"


class Map(BaseAnyWidget):
    """
    The top-level class used to display a map in a Jupyter Widget.

    **Example:**

    ```py
    import geopandas as gpd
    from lonboard import Map, ScatterplotLayer, SolidPolygonLayer

    # A GeoDataFrame with Point geometries
    point_gdf = gpd.GeoDataFrame()
    point_layer = ScatterplotLayer.from_geopandas(
        point_gdf,
        get_fill_color=[255, 0, 0],
    )

    # A GeoDataFrame with Polygon geometries
    polygon_gdf = gpd.GeoDataFrame()
    polygon_layer = SolidPolygonLayer.from_geopandas(
        gdf,
        get_fill_color=[255, 0, 0],
    )

    m = Map([point_layer, polygon_layer])
    ```
    """

    def __init__(self, layers: Union[BaseLayer, Sequence[BaseLayer]], **kwargs) -> None:
        """Create a new Map.

        Aside from the `layers` argument, pass keyword arguments for any other attribute
        defined in this class.

        Args:
            layers: One or more layers to render on this map.

        Returns:
            A Map object.
        """
        if isinstance(layers, BaseLayer):
            layers = [layers]

        super().__init__(layers=layers, **kwargs)

    _esm = bundler_output_dir / "index.js"
    _css = bundler_output_dir / "index.css"

    _initial_view_state = traitlets.Dict().tag(sync=True)
    """
    The initial view state of the map.

    - Type: `dict`, optional
    - Default: Automatically inferred from the data passed to the map.

    The keys _must_ include:

    - `longitude`: longitude at the map center.
    - `latitude`: latitude at the map center.
    - `zoom`: zoom level.

    Keys may additionally include:

    - `pitch` (float, optional) - pitch angle in degrees. Default `0` (top-down).
    - `bearing` (float, optional) - bearing angle in degrees. Default `0` (north).
    - `maxZoom` (float, optional) - max zoom level. Default `20`.
    - `minZoom` (float, optional) - min zoom level. Default `0`.
    - `maxPitch` (float, optional) - max pitch angle. Default `60`.
    - `minPitch` (float, optional) - min pitch angle. Default `0`.

    Note that currently no camel-case/snake-case translation occurs for this method, and
    so keys must be in camel case.

    This API is not yet stabilized and may change in the future.
    """

    _height = traitlets.Int(default_value=DEFAULT_HEIGHT, allow_none=True).tag(
        sync=True
    )
    """Height of the map in pixels.

    This API is not yet stabilized and may change in the future.
    """

    layers = traitlets.List(trait=traitlets.Instance(BaseLayer)).tag(
        sync=True, **ipywidgets.widget_serialization
    )
    """One or more `Layer` objects to display on this map.
    """

    show_tooltip = traitlets.Bool(True).tag(sync=True)
    """
    Whether to render a tooltip on hover on the map.

    - Type: `bool`
    - Default: `True`
    """

    picking_radius = traitlets.Int(5).tag(sync=True)
    """
    Extra pixels around the pointer to include while picking (such as for a tooltip).

    This is helpful when rendered objects are difficult to target, for example
    irregularly shaped icons, small moving circles or interaction by touch.

    - Type: `int`
    - Default: `5`
    """

    basemap_style = traitlets.Unicode(CartoBasemap.PositronNoLabels).tag(sync=True)
    """
    A MapLibre-compatible basemap style.

    Various styles are provided in [`lonboard.basemap`](https://developmentseed.org/lonboard/latest/api/basemap/).

    - Type: `str`
    - Default
      [`lonboard.basemap.CartoBasemap.PositronNoLabels`][lonboard.basemap.CartoBasemap.PositronNoLabels]
    """

    def to_html(self, filename: Union[str, Path]) -> None:
        """Save the current map as a standalone HTML file.

        Args:
            filename: where to save the generated HTML file.
        """
        embed_minimal_html(filename, views=[self], drop_defaults=False)

    @traitlets.default("_initial_view_state")
    def _default_initial_view_state(self):
        return compute_view(self.layers)



"""Reproject a GeoArrow array
"""
import json
import warnings
from concurrent.futures import ThreadPoolExecutor
from functools import lru_cache, partial
from typing import Callable, Optional, Tuple, Union

import numpy as np
import pyarrow as pa
from pyproj import CRS, Transformer

from lonboard._constants import EXTENSION_NAME, OGC_84
from lonboard._geoarrow.extension_types import CoordinateDimension
from lonboard._utils import get_geometry_column_index

TransformerFromCRS = lru_cache(Transformer.from_crs)


def reproject_table(
    table: pa.Table,
    *,
    to_crs: Union[str, CRS] = OGC_84,
    max_workers: Optional[int] = None,
) -> pa.Table:
    """Reproject a GeoArrow table to a new CRS

    Args:
        table: The table to reproject.
        to_crs: The target CRS. Defaults to OGC_84.
        max_workers: The maximum number of threads to use. Defaults to None.

    Returns:
        A new table.
    """
    geom_col_idx = get_geometry_column_index(table.schema)
    # No geometry column in table
    if geom_col_idx is None:
        return table

    geom_field = table.schema.field(geom_col_idx)
    geom_column = table.column(geom_col_idx)

    # geometry column exists in table but is not assigned a CRS
    if b"ARROW:extension:metadata" not in geom_field.metadata:
        return table

    new_field, new_column = reproject_column(
        field=geom_field, column=geom_column, to_crs=to_crs, max_workers=max_workers
    )
    return table.set_column(geom_col_idx, new_field, new_column)


def reproject_column(
    *,
    field: pa.Field,
    column: pa.ChunkedArray,
    to_crs: Union[str, CRS] = OGC_84,
    max_workers: Optional[int] = None,
) -> Tuple[pa.Field, pa.ChunkedArray]:
    """Reproject a GeoArrow array to a new CRS

    Args:
        field: The field describing the column
        column: A ChunkedArray
        to_crs: The target CRS. Defaults to OGC_84.
        max_workers: The maximum number of threads to use. Defaults to None.
    """
    extension_type_name = field.metadata[b"ARROW:extension:name"]
    extension_metadata = json.loads(field.metadata[b"ARROW:extension:metadata"])
    crs_str = extension_metadata["crs"]
    existing_crs = CRS(crs_str)

    if existing_crs == to_crs:
        return field, column

    # NOTE: Not sure the best place to put this warning
    warnings.warn("Input being reprojected to EPSG:4326 CRS")

    transformer = TransformerFromCRS(existing_crs, to_crs, always_xy=True)

    # Metadata inside metadata, bad naming
    new_extension_meta_meta = {"crs": CRS(to_crs).to_json()}
    new_extension_metadata = {
        b"ARROW:extension:name": extension_type_name,
        b"ARROW:extension:metadata": json.dumps(new_extension_meta_meta),
    }

    new_chunked_array = _reproject_column(
        column,
        extension_type_name=extension_type_name,
        transformer=transformer,
        max_workers=max_workers,
    )
    return field.with_metadata(new_extension_metadata), new_chunked_array


def _reproject_column(
    column: pa.ChunkedArray,
    *,
    extension_type_name: EXTENSION_NAME,
    transformer: Transformer,
    max_workers: Optional[int] = None,
) -> pa.ChunkedArray:
    if extension_type_name == EXTENSION_NAME.POINT:
        func = partial(_reproject_chunk_nest_0, transformer=transformer)
    elif extension_type_name in [EXTENSION_NAME.LINESTRING, EXTENSION_NAME.MULTIPOINT]:
        func = partial(_reproject_chunk_nest_1, transformer=transformer)
    elif extension_type_name in [
        EXTENSION_NAME.POLYGON,
        EXTENSION_NAME.MULTILINESTRING,
    ]:
        func = partial(_reproject_chunk_nest_2, transformer=transformer)

    elif extension_type_name == EXTENSION_NAME.MULTIPOLYGON:
        func = partial(_reproject_chunk_nest_3, transformer=transformer)
    else:
        raise ValueError(f"Unexpected extension type name {extension_type_name}")

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        return pa.chunked_array(executor.map(func, column.chunks))


def _reproject_coords(arr: pa.FixedSizeListArray, transformer: Transformer):
    list_size = arr.type.list_size
    np_arr = arr.flatten().to_numpy().reshape(-1, list_size)

    if list_size == 2:
        output_np_arr = np.column_stack(
            transformer.transform(np_arr[:, 0], np_arr[:, 1])
        )
        dims = CoordinateDimension.XY
    elif list_size == 3:
        output_np_arr = np.column_stack(
            transformer.transform(np_arr[:, 0], np_arr[:, 1], np_arr[:, 2])
        )
        dims = CoordinateDimension.XYZ
    else:
        raise ValueError(f"Unexpected list size {list_size}")

    coord_field = pa.list_(pa.field(dims, pa.float64()), len(dims))
    return pa.FixedSizeListArray.from_arrays(
        output_np_arr.flatten("C"), type=coord_field
    )


def _reproject_chunk_nest_0(arr: pa.ListArray, transformer: Transformer):
    callback = partial(_reproject_coords, transformer=transformer)
    return _map_coords_nest_0(arr, callback)


def _reproject_chunk_nest_1(arr: pa.ListArray, transformer: Transformer):
    callback = partial(_reproject_coords, transformer=transformer)
    return _map_coords_nest_1(arr, callback)


def _reproject_chunk_nest_2(arr: pa.ListArray, transformer: Transformer):
    callback = partial(_reproject_coords, transformer=transformer)
    return _map_coords_nest_2(arr, callback)


def _reproject_chunk_nest_3(arr: pa.ListArray, transformer: Transformer):
    callback = partial(_reproject_coords, transformer=transformer)
    return _map_coords_nest_3(arr, callback)


def _map_coords_nest_0(
    arr: pa.FixedSizeListArray,
    callback: Callable[[pa.FixedSizeListArray], pa.FixedSizeListArray],
):
    new_coords = callback(arr)
    return new_coords


def _map_coords_nest_1(
    arr: pa.ListArray,
    callback: Callable[[pa.FixedSizeListArray], pa.FixedSizeListArray],
):
    geom_offsets = arr.offsets
    coords = arr.flatten()
    new_coords = callback(coords)
    new_geometry_array = pa.ListArray.from_arrays(geom_offsets, new_coords)
    return new_geometry_array


def _map_coords_nest_2(
    arr: pa.ListArray,
    callback: Callable[[pa.FixedSizeListArray], pa.FixedSizeListArray],
):
    geom_offsets = arr.offsets
    ring_offsets = arr.flatten().offsets
    coords = arr.flatten().flatten()
    new_coords = callback(coords)
    new_ring_array = pa.ListArray.from_arrays(ring_offsets, new_coords)
    new_geometry_array = pa.ListArray.from_arrays(geom_offsets, new_ring_array)
    return new_geometry_array


def _map_coords_nest_3(
    arr: pa.ListArray,
    callback: Callable[[pa.FixedSizeListArray], pa.FixedSizeListArray],
):
    geom_offsets = arr.offsets
    polygon_offsets = arr.flatten().offsets
    ring_offsets = arr.flatten().flatten().offsets
    coords = arr.flatten().flatten().flatten()
    new_coords = callback(coords)
    new_ring_array = pa.ListArray.from_arrays(ring_offsets, new_coords)
    new_polygon_array = pa.ListArray.from_arrays(polygon_offsets, new_ring_array)
    new_geometry_array = pa.ListArray.from_arrays(geom_offsets, new_polygon_array)
    return new_geometry_array



import math
from io import BytesIO
from typing import List, Tuple, Union

import numpy as np
import pyarrow as pa
import pyarrow.parquet as pq
from numpy.typing import NDArray
from traitlets import TraitError

DEFAULT_PARQUET_COMPRESSION = "ZSTD"
DEFAULT_PARQUET_COMPRESSION_LEVEL = 7
DEFAULT_PARQUET_CHUNK_SIZE = 2**16
# Target chunk size for Arrow (uncompressed) per Parquet chunk
DEFAULT_ARROW_CHUNK_BYTES_SIZE = 5 * 1024 * 1024  # 5MB

# Maximum number of separate chunks/row groups to allow splitting an input layer into
# Deck.gl can pick from a maximum of 256 layers, and a user could have many layers, so
# we don't want to use too many layers per data file.
DEFAULT_MAX_NUM_CHUNKS = 32


def serialize_table_to_parquet(table: pa.Table, *, max_chunksize: int) -> List[bytes]:
    buffers: List[bytes] = []
    # NOTE: passing `max_chunksize=0` creates an infinite loop
    # https://github.com/apache/arrow/issues/39788
    assert max_chunksize > 0

    for record_batch in table.to_batches(max_chunksize=max_chunksize):
        with BytesIO() as bio:
            with pq.ParquetWriter(
                bio,
                schema=table.schema,
                compression=DEFAULT_PARQUET_COMPRESSION,
                compression_level=DEFAULT_PARQUET_COMPRESSION_LEVEL,
            ) as writer:
                # Occasionally it's possible for there to be empty batches in the
                # pyarrow table. This will error when writing to parquet. We want to
                # give a more informative error.
                if record_batch.num_rows == 0:
                    raise ValueError("Batch with 0 rows.")

                writer.write_batch(record_batch, row_group_size=record_batch.num_rows)

            buffers.append(bio.getvalue())

    return buffers


def serialize_pyarrow_column(data: pa.Array, *, max_chunksize: int) -> List[bytes]:
    """Serialize a pyarrow column to a Parquet file with one column"""
    pyarrow_table = pa.table({"value": data})
    return serialize_table_to_parquet(pyarrow_table, max_chunksize=max_chunksize)


def serialize_accessor(data: Union[List[int], Tuple[int], NDArray[np.uint8]], obj):
    if data is None:
        return None

    # We assume data has already been validated to the right type for this accessor
    # Allow any json-serializable type through
    if isinstance(data, (str, int, float, list, tuple, bytes)):
        return data

    assert isinstance(data, (pa.ChunkedArray, pa.Array))
    validate_accessor_length_matches_table(data, obj.table)
    return serialize_pyarrow_column(data, max_chunksize=obj._rows_per_chunk)


def serialize_table(data, obj):
    assert isinstance(data, pa.Table), "expected pyarrow table"
    return serialize_table_to_parquet(data, max_chunksize=obj._rows_per_chunk)


def infer_rows_per_chunk(table: pa.Table) -> int:
    # At least one chunk
    num_chunks = max(round(table.nbytes / DEFAULT_ARROW_CHUNK_BYTES_SIZE), 1)

    # Clamp to the maximum number of chunks
    num_chunks = min(num_chunks, DEFAULT_MAX_NUM_CHUNKS)

    rows_per_chunk = math.ceil((table.num_rows / num_chunks))
    return rows_per_chunk


def validate_accessor_length_matches_table(accessor, table):
    if len(accessor) != len(table):
        raise TraitError("accessor must have same length as table")


ACCESSOR_SERIALIZATION = {"to_json": serialize_accessor}
TABLE_SERIALIZATION = {"to_json": serialize_table}



"""Remove custom geoarrow.pyarrow types from input geoarrow data
"""
import json
from typing import Tuple

import pyarrow as pa
from pyproj import CRS


def sanitize_table(table: pa.Table) -> pa.Table:
    """
    Convert any registered geoarrow.pyarrow extension fields and arrays to plain
    metadata
    """
    for field_idx in range(len(table.schema)):
        field = table.field(field_idx)
        column = table.column(field_idx)

        if isinstance(field.type, pa.ExtensionType):
            assert all(isinstance(chunk, pa.ExtensionArray) for chunk in column.chunks)
            new_field, new_column = sanitize_column(field, column)
            table = table.set_column(field_idx, new_field, new_column)

    return table


def sanitize_column(
    field: pa.Field, column: pa.ChunkedArray
) -> Tuple[pa.Field, pa.ChunkedArray]:
    """
    Convert a registered geoarrow.pyarrow extension field and column to plain metadata
    """
    import geoarrow.pyarrow as gap

    extension_metadata = {}
    if field.type.crs:
        extension_metadata["crs"] = CRS.from_user_input(field.type.crs).to_json()

    if field.type.edge_type == gap.EdgeType.SPHERICAL:
        extension_metadata["edges"] = "spherical"

    metadata = {
        "ARROW:extension:name": field.type.extension_name,
    }
    if extension_metadata:
        metadata["ARROW:extension:metadata"] = json.dumps(extension_metadata)

    new_field = pa.field(
        field.name, field.type.storage_type, nullable=field.nullable, metadata=metadata
    )

    new_chunks = []
    for chunk in column.chunks:
        if hasattr(chunk, "storage"):
            new_chunks.append(chunk.storage)
        else:
            new_chunks.append(chunk.cast(new_field.type))

    return new_field, pa.chunked_array(new_chunks)



import traitlets

from lonboard._base import BaseExtension
from lonboard.experimental.traits import GetFilterValueAccessor, PointAccessor
from lonboard.traits import FloatAccessor


class BrushingExtension(BaseExtension):
    """
    Adds GPU-based data brushing functionalities to layers. It allows the layer to
    show/hide objects based on the current pointer position.

    # Example

    An example is in the [County-to-County Migration
    notebook](https://developmentseed.org/lonboard/latest/examples/migration/).

    # Layer Properties

    This extension dynamically enables the following properties onto the layer(s) where
    it is included:

    ## `brushing_enabled`

    Enable/disable brushing. If brushing is disabled, all objects are rendered.

    - Type: `bool`, optional
    - Default: `True`

    ## `brushing_target`

    The position used to filter each object by. One of the following:

    - `"source"`: Use the primary position for each object. This can mean different
      things depending on the layer. It usually refers to the coordinates returned by
      `getPosition` or `getSourcePosition` accessors.
    - `"target"`: Use the secondary position for each object. This may not be available
      in some layers. It usually refers to the coordinates returned by
      `getTargetPosition` accessor.
    - `"source_target"`: Use both the primary position and secondary position for each
      object. Show object if either is in brushing range.
    - `"custom"`: Some layers may not describe their data objects with one or two
      coordinates, for example `PathLayer` and `PolygonLayer`. Use this option with the
      `get_brushing_target` prop to provide a custom position that each object should be
      filtered by.

    - Type: `str`, optional

        One of: "source" | "target" | "source_target" | "custom"

    - Default: `"source"`

    ## `brushing_radius`

    The brushing radius centered at the pointer, in meters. If a data object is within
    this circle, it is rendered; otherwise it is hidden.

    - Type: `float`, optional
    - Default: `10000`

    ## `get_brushing_target`

    An arbitrary position for each object that it will be filtered by.

    Only effective if `brushing_target` is set to `"custom"`.

    - Type: [PointAccessor][lonboard.experimental.traits.PointAccessor], optional
        - If a point is provided, it is used as the target for all rows.
        - If an array of points is provided, each value in the array will be used as the
          target for the row at the same row index.
    - Default: `None`.
    """

    _extension_type = traitlets.Unicode("brushing").tag(sync=True)

    _layer_traits = {
        "brushing_enabled": traitlets.Bool(True).tag(sync=True),
        "brushing_target": traitlets.Unicode(None, allow_none=True).tag(sync=True),
        "brushing_radius": traitlets.Float(None, allow_none=True, min=0).tag(sync=True),
        "get_brushing_target": PointAccessor(None, allow_none=True),
    }


class CollisionFilterExtension(BaseExtension):
    """Allows layers to hide overlapping objects.

    # Layer Properties

    This extension dynamically enables the following properties onto the layer(s) where
    it is included:

    ## `collision_enabled`

    Enable/disable collisions. If collisions are disabled, all objects are rendered.

    - Type: `bool`, optional
    - Default: `True`

    ## `collision_group`

    Collision group this layer belongs to. If it is not set, the 'default' collision
    group is used

    - Type: `str`, optional
    - Default: `None`

    ## `get_collision_priority`

    Accessor for collision priority. Must return a number in the range -1000 -> 1000.
    Features with higher values are shown preferentially.

    - Type: [FloatAccessor][lonboard.traits.FloatAccessor], optional
        - If a number is provided, it is used as the priority for all objects.
        - If an array is provided, each value in the array will be used as the priority
          for the object at the same row index.
    - Default: `0`.

    """

    _extension_type = traitlets.Unicode("collision-filter").tag(sync=True)

    _layer_traits = {
        "collision_enabled": traitlets.Bool(True).tag(sync=True),
        "collision_group": traitlets.Unicode(None, allow_none=True).tag(sync=True),
        "get_collision_priority": FloatAccessor(None, allow_none=True),
    }


class DataFilterExtension(BaseExtension):
    """
    Adds GPU-based data filtering functionalities to layers. It allows the layer to
    show/hide objects based on user-defined properties.

    # Example

    ```py
    from lonboard import Map, ScatterplotLayer
    from lonboard.colormap import apply_continuous_cmap
    from lonboard.experimental import DataFilterExtension

    gdf = gpd.GeoDataFrame(...)
    extension = DataFilterExtension()
    layer = ScatterplotLayer.from_geopandas(
        gdf,
        extensions=[extension],
        get_filter_value=gdf["filter_value"], # replace with desired column
        filter_range=[0, 5] # replace with desired filter range
    )
    ```

    The `DataFilterExtension` allows filtering on 1 to 4 attributes at the same time. So
    if you have four numeric columns of interest, you can filter on the intersection of
    all of them.

    For easy visualization, we suggest connecting the `DataFilterExtension` to an
    interactive slider from `ipywidgets`.

    ```py
    from ipywidgets import FloatRangeSlider

    slider = FloatRangeSlider(
        value=(2, 5),
        min=0,
        max=10,
        step=0.1,
        description="Slider: "
    )
    slider

    jsdlink(
        (slider, "value"),
        (layer, "filter_range")
    )
    ```

    If you have 2 to 4 columns, use a
    [`MultiRangeSlider`][lonboard.controls.MultiRangeSlider], which combines multiple
    `FloatRangeSlider` objects in a form that the `DataFilterExtension` expects.

    ```py
    from ipywidgets import FloatRangeSlider, jsdlink

    slider1 = FloatRangeSlider(
        value=(2, 5),
        min=0,
        max=10,
        step=0.1,
        description="First slider: "
    )
    slider2 = FloatRangeSlider(
        value=(30, 40),
        min=0,
        max=50,
        step=1,
        description="Second slider: "
    )
    multi_slider = MultiRangeSlider([slider1, slider2])
    multi_slider

    jsdlink(
        (multi_slider, "value"),
        (layer, "filter_range")
    )
    ```

    # Important notes

    - The DataFilterExtension only supports float32 data, so integer data will be casted
      to float32.
    - The DataFilterExtension copies all data referenced by `get_filter_value` to the
      GPU, so it will increase memory pressure on the GPU.

    # Layer Properties

    ## `filter_enabled`

    Enable/disable the data filter. If the data filter is disabled, all objects are
    rendered.

    - Type: `bool`, optional
    - Default: `True`

    ## `filter_range`

    The bounds which defines whether an object should be rendered. If an object's
    filtered value is within the bounds, the object will be rendered; otherwise it will
    be hidden. This prop can be updated on user input or animation with very little
    cost.

    Format:

    If `filter_size` is 1, provide a single tuple of `(min, max)`.

    If `filter_size` is 2 to 4, provide a list of tuples: `[(min0, max0), (min1,
    max1), ...]` for each filtered property, respectively.

    - Type: either Tuple[float, float] or List[Tuple[float, float]], optional
    - Default: `(-1, 1)`

    ## `filter_soft_range`

    If specified, objects will be faded in/out instead of abruptly shown/hidden.

    When the filtered value is outside of the bounds defined by `filter_soft_range` but
    still within the bounds defined by `filter_range`, the object will be rendered as
    "faded".

    - Type: Tuple[float, float], optional
    - Default: `None`

    ## `filter_transform_size`

    When an object is "faded", manipulate its size so that it appears smaller or
    thinner. Only works if `filter_soft_range` is specified.

    - Type: `bool`, optional
    - Default: `True`

    ## `filter_transform_color`

    When an object is "faded", manipulate its opacity so that it appears more
    translucent. Only works if `filter_soft_range` is specified.

    - Type: `bool`, optional
    - Default: `True`

    ## `get_filter_value`

    Accessor to retrieve the value for each object that it will be filtered by.

    - Type:
      [GetFilterValueAccessor][lonboard.experimental.traits.GetFilterValueAccessor]
        - If a scalar value is provided, it is used as the value for all objects.
        - If an array is provided, each value in the array will be used as the value
          for the object at the same row index.
    """

    _extension_type = traitlets.Unicode("data-filter").tag(sync=True)

    _layer_traits = {
        "filter_enabled": traitlets.Bool(True).tag(sync=True),
        "filter_range": traitlets.Union(
            [
                traitlets.List(traitlets.Float(), minlen=2, maxlen=2),
                traitlets.List(
                    traitlets.List(traitlets.Float(), minlen=2, maxlen=2),
                    minlen=2,
                    maxlen=4,
                ),
            ]
        ).tag(sync=True),
        "filter_soft_range": traitlets.Tuple(
            traitlets.Float(), traitlets.Float(), default_value=None, allow_none=True
        ).tag(sync=True),
        "filter_transform_size": traitlets.Bool(True).tag(sync=True),
        "filter_transform_color": traitlets.Bool(True).tag(sync=True),
        "get_filter_value": GetFilterValueAccessor(None, allow_none=False),
    }

    filter_size = traitlets.Int(1, min=1, max=4).tag(sync=True)
    """The size of the filter (number of columns to filter by).

    The data filter can show/hide data based on 1-4 numeric properties of each object.

    - Type: `int`, optional
    - Default 1.
    """



from __future__ import annotations

from typing import Any, List, Tuple, Union

import matplotlib as mpl
import numpy as np
import pyarrow as pa
from traitlets.traitlets import TraitType

from lonboard._serialization import ACCESSOR_SERIALIZATION
from lonboard.traits import FixedErrorTraitType


class PointAccessor(FixedErrorTraitType):
    """A representation of a deck.gl point accessor.

    Various input is allowed:

    - A numpy `ndarray` with two dimensions and data type [`np.uint8`][numpy.uint8]. The
      size of the second dimension must be `2` or `3`, and will correspond to either XY
      or XYZ positions.
    - A pyarrow [`FixedSizeListArray`][pyarrow.FixedSizeListArray] or
      [`ChunkedArray`][pyarrow.ChunkedArray] containing `FixedSizeListArray`s. The inner
      size of the fixed size list must be `2` or `3` and its child must be of floating
      point type.
    """

    default_value = (0, 0, 0)
    info_text = (
        "a tuple or list representing an RGB(A) color or numpy ndarray or "
        "pyarrow FixedSizeList representing an array of RGB(A) colors"
    )

    def __init__(
        self: TraitType,
        *args,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def validate(
        self, obj, value
    ) -> Union[Tuple[int, ...], List[int], pa.ChunkedArray, pa.FixedSizeListArray]:
        if isinstance(value, np.ndarray):
            if value.ndim != 2:
                self.error(obj, value, info="Point array to have 2 dimensions")

            list_size = value.shape[1]
            if list_size not in (2, 3):
                self.error(
                    obj,
                    value,
                    info="Point array to have 2 or 3 as its second dimension",
                )

            return pa.FixedSizeListArray.from_arrays(value.flatten("C"), list_size)

        if isinstance(value, (pa.ChunkedArray, pa.Array)):
            if not pa.types.is_fixed_size_list(value.type):
                self.error(obj, value, info="Point pyarrow array to be a FixedSizeList")

            if value.type.list_size not in (2, 3):
                self.error(
                    obj,
                    value,
                    info=(
                        "Color pyarrow array to be a FixedSizeList with list size of "
                        "2 or 3"
                    ),
                )

            if not pa.types.is_floating(value.type.value_type):
                self.error(
                    obj,
                    value,
                    info="Point pyarrow array to have a floating point child",
                )

            return value

        if isinstance(value, str):
            try:
                c = mpl.colors.to_rgba(value)  # type: ignore
            except ValueError:
                self.error(
                    obj,
                    value,
                    info=(
                        "Color string to be a hex string interpretable by "
                        "matplotlib.colors.to_rgba"
                    ),
                )
                return

            return tuple(map(int, (np.array(c) * 255).astype(np.uint8)))

        self.error(obj, value)
        assert False


class GetFilterValueAccessor(FixedErrorTraitType):
    """
    A trait to validate input for the `get_filter_value` accessor added by the
    [`DataFilterExtension`][lonboard.experimental.DataFilterExtension], which can have
    between 1 and 4 float values per row.

    Various input is allowed:

    - An `int` or `float`. This will be used as the value for all objects. The
      `filter_size` of the
      [`DataFilterExtension`][lonboard.experimental.DataFilterExtension] instance must
      be 1.
    - A one-dimensional numpy `ndarray` with a numeric data type. This will be casted to
      an array of data type [`np.float32`][numpy.float32]. Each value in the array will
      be used as the value for the object at the same row index. The `filter_size` of
      the [`DataFilterExtension`][lonboard.experimental.DataFilterExtension] instance
      must be 1.
    - A two-dimensional numpy `ndarray` with a numeric data type. This will be casted to
      an array of data type [`np.float32`][numpy.float32]. Each value in the array will
      be used as the value for the object at the same row index. The `filter_size` of
      the [`DataFilterExtension`][lonboard.experimental.DataFilterExtension] instance
      must match the size of the second dimension of the array.
    - A pandas `Series` with a numeric data type. This will be casted to an array of
      data type [`np.float32`][numpy.float32]. Each value in the array will be used as
      the value for the object at the same row index. The `filter_size` of
      the [`DataFilterExtension`][lonboard.experimental.DataFilterExtension] instance
      must be 1.
    - A pyarrow [`FloatArray`][pyarrow.FloatArray], [`DoubleArray`][pyarrow.DoubleArray]
      or [`ChunkedArray`][pyarrow.ChunkedArray] containing either a `FloatArray` or
      `DoubleArray`. Each value in the array will be used as the value for the object at
      the same row index. The `filter_size` of the
      [`DataFilterExtension`][lonboard.experimental.DataFilterExtension] instance must
      be 1.

      Alternatively, you can pass any corresponding Arrow data structure from a library
      that implements the [Arrow PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
    - A pyarrow [`FixedSizeListArray`][pyarrow.FixedSizeListArray] or
      [`ChunkedArray`][pyarrow.ChunkedArray] containing `FixedSizeListArray`s. The child
      array of the fixed size list must be of floating point type. The `filter_size` of
      the [`DataFilterExtension`][lonboard.experimental.DataFilterExtension] instance
      must match the list size.

      Alternatively, you can pass any corresponding Arrow data structure from a library
      that implements the [Arrow PyCapsule
      Interface](https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html).
    """

    default_value = float(0)
    info_text = (
        "a float value or numpy ndarray or pyarrow array representing an array"
        " of floats"
    )

    def __init__(
        self: TraitType,
        *args,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.tag(sync=True, **ACCESSOR_SERIALIZATION)

    def validate(self, obj, value) -> Union[float, pa.ChunkedArray, pa.DoubleArray]:
        # Find the data filter extension in the attributes of the parent object so we
        # can validate against the filter size.
        data_filter_extension = [
            ext for ext in obj.extensions if ext._extension_type == "data-filter"
        ]
        assert len(data_filter_extension) == 1
        filter_size = data_filter_extension[0].filter_size

        if isinstance(value, (int, float)):
            if filter_size != 1:
                self.error(obj, value, info="filter_size==1 with scalar value")

            return float(value)

        if isinstance(value, (tuple, list)):
            if filter_size != len(value):
                self.error(
                    obj,
                    value,
                    info=f"filter_size ({filter_size}) to match length of tuple/list",
                )

            if any(not isinstance(v, (int, float)) for v in value):
                self.error(
                    obj,
                    value,
                    info="all values in tuple or list to be numeric",
                )

            return value

        # pandas Series
        if (
            value.__class__.__module__.startswith("pandas")
            and value.__class__.__name__ == "Series"
        ):
            # Assert that filter_size == 1 for a pandas series.
            # Pandas series can technically contain Python list objects inside them, but
            # for simplicity we disallow that.
            if filter_size != 1:
                self.error(obj, value, info="filter_size==1 with pandas Series")

            # Cast pandas Series to numpy ndarray
            value = np.asarray(value)

        if isinstance(value, np.ndarray):
            if not np.issubdtype(value.dtype, np.number):
                self.error(obj, value, info="numeric dtype")

            # Cast to float32
            value = value.astype(np.float32)

            if len(value.shape) == 1:
                if filter_size != 1:
                    self.error(obj, value, info="filter_size==1 with 1-D numpy array")

                return pa.array(value)

            if len(value.shape) != 2:
                self.error(obj, value, info="1-D or 2-D numpy array")

            if value.shape[1] != filter_size:
                self.error(
                    obj,
                    value,
                    info=(
                        f"filter_size ({filter_size}) to match 2nd dimension of "
                        "numpy array"
                    ),
                )

            return pa.FixedSizeListArray.from_arrays(value.flatten("C"), filter_size)

        # Check for Arrow PyCapsule Interface
        # https://arrow.apache.org/docs/format/CDataInterface/PyCapsuleInterface.html
        # TODO: with pyarrow v16 also import chunked array from stream
        if not isinstance(value, (pa.ChunkedArray, pa.Array)):
            if hasattr(value, "__arrow_c_array__"):
                value = pa.array(value)

        if isinstance(value, (pa.ChunkedArray, pa.Array)):
            # Allowed inputs are either a FixedSizeListArray or numeric array.
            # If not a fixed size list array, check for floating and cast to float32
            if not pa.types.is_fixed_size_list(value.type):
                if filter_size != 1:
                    self.error(
                        obj,
                        value,
                        info="filter_size==1 with non-FixedSizeList type arrow array",
                    )

                if not pa.types.is_floating(value.type):
                    self.error(
                        obj,
                        value,
                        info="arrow array to be a floating point type",
                    )

                return value.cast(pa.float32())

            # We have a FixedSizeListArray
            if filter_size != value.type.list_size:
                self.error(
                    obj,
                    value,
                    info=(
                        f"filter_size ({filter_size}) to match list size of "
                        "FixedSizeList arrow array"
                    ),
                )

            if not pa.types.is_floating(value.type.value_type):
                self.error(
                    obj,
                    value,
                    info="arrow array to have floating point child type",
                )

            # Cast values to float32
            return value.cast(pa.list_(pa.float32(), value.type.list_size))

        self.error(obj, value)
        assert False



from typing import Dict

import traitlets
from anywidget import AnyWidget
from ipywidgets import Widget


class BaseWidget(Widget):
    def __init__(self, **kwargs):
        # Raise error on unknown keyword name
        # Note: we don't use `class_own_traits()` because some layer props are set on
        # BaseLayer
        layer_trait_names = self.trait_names()
        for provided_trait_name in kwargs.keys():
            if provided_trait_name not in layer_trait_names:
                raise TypeError(f"unexpected keyword argument '{provided_trait_name}'")

        super().__init__(**kwargs)


class BaseAnyWidget(AnyWidget):
    def __init__(self, **kwargs):
        # Raise error on unknown keyword name
        layer_trait_names = self.trait_names()
        for provided_trait_name in kwargs.keys():
            if provided_trait_name not in layer_trait_names:
                raise TypeError(f"unexpected keyword argument '{provided_trait_name}'")

        super().__init__(**kwargs)


class BaseExtension(BaseWidget):
    _layer_traits: Dict[str, traitlets.TraitType] = {}
    """Traits from this extension to dynamically assign onto a layer."""



import json
from enum import Enum
from typing import Dict, Optional, Tuple

import numpy as np
import pyarrow as pa
import shapely
from numpy.typing import NDArray
from shapely import GeometryType


class CoordinateDimension(str, Enum):
    XY = "xy"
    XYZ = "xyz"
    XYM = "xym"
    XYZM = "xyzm"


class BaseGeometryType(pa.ExtensionType):
    extension_name: str
    coord_dimension: CoordinateDimension


def coord_storage_type(*, interleaved: bool, dims: CoordinateDimension) -> pa.DataType:
    """Generate the storage type of a geoarrow coordinate array

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
    """
    if interleaved:
        return pa.list_(pa.field(dims, pa.float64()), len(dims))

    else:
        if dims == CoordinateDimension.XY:
            return pa.struct(
                [
                    ("x", pa.float64()),
                    ("y", pa.float64()),
                ]
            )
        if dims == CoordinateDimension.XYZ:
            return pa.struct(
                [
                    ("x", pa.float64()),
                    ("y", pa.float64()),
                    ("z", pa.float64()),
                ]
            )
        if dims == CoordinateDimension.XYM:
            return pa.struct(
                [
                    ("x", pa.float64()),
                    ("y", pa.float64()),
                    ("m", pa.float64()),
                ]
            )
        if dims == CoordinateDimension.XYZM:
            return pa.struct(
                [
                    ("x", pa.float64()),
                    ("y", pa.float64()),
                    ("z", pa.float64()),
                    ("m", pa.float64()),
                ]
            )


def linestring_storage_type(
    *, interleaved: bool, dims: CoordinateDimension, large_list: bool = False
) -> pa.DataType:
    """Generate the storage type of a geoarrow.linestring array

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
        large_list: Whether to use a large list with int64 offsets for the inner type
    """
    vertices_type = coord_storage_type(interleaved=interleaved, dims=dims)
    if large_list:
        return pa.large_list(pa.field("vertices", vertices_type))
    else:
        return pa.list_(pa.field("vertices", vertices_type))


def polygon_storage_type(
    *, interleaved: bool, dims: CoordinateDimension, large_list: bool = False
) -> pa.DataType:
    """Generate the storage type of a geoarrow.polygon array

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
        large_list: Whether to use a large list with int64 offsets for the inner type
    """
    rings_type = linestring_storage_type(
        large_list=large_list, interleaved=interleaved, dims=dims
    )
    if large_list:
        return pa.large_list(pa.field("rings", rings_type))
    else:
        return pa.list_(pa.field("rings", rings_type))


def multipoint_storage_type(
    *, interleaved: bool, dims: CoordinateDimension, large_list: bool = False
) -> pa.DataType:
    """Generate the storage type of a geoarrow.multipoint array

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
        large_list: Whether to use a large list with int64 offsets for the inner type
    """
    points_type = coord_storage_type(interleaved=interleaved, dims=dims)
    if large_list:
        return pa.large_list(pa.field("points", points_type))
    else:
        return pa.list_(pa.field("points", points_type))


def multilinestring_storage_type(
    *, interleaved: bool, dims: CoordinateDimension, large_list: bool = False
) -> pa.DataType:
    """Generate the storage type of a geoarrow.multilinestring array

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
        large_list: Whether to use a large list with int64 offsets for the inner type
    """
    linestrings_type = linestring_storage_type(
        large_list=large_list, interleaved=interleaved, dims=dims
    )
    if large_list:
        return pa.large_list(pa.field("linestrings", linestrings_type))
    else:
        return pa.list_(pa.field("linestrings", linestrings_type))


def multipolygon_storage_type(
    *, interleaved: bool, dims: CoordinateDimension, large_list: bool = False
) -> pa.DataType:
    """Generate the storage type of a geoarrow.multipolygon array

    Args:
        interleaved: Whether coordinates should be interleaved or separated
        dims: The number of dimensions
        large_list: Whether to use a large list with int64 offsets for the inner type
    """
    polygons_type = polygon_storage_type(
        large_list=large_list, interleaved=interleaved, dims=dims
    )
    if large_list:
        return pa.large_list(pa.field("polygons", polygons_type))
    else:
        return pa.list_(pa.field("polygons", polygons_type))


class PointType(BaseGeometryType):
    extension_name = "geoarrow.point"

    def __init__(self, *, interleaved: bool, dims: CoordinateDimension):
        self.coord_dimension = dims

        storage_type = coord_storage_type(interleaved=interleaved, dims=dims)
        super().__init__(storage_type, self.extension_name)

    def __arrow_ext_serialize__(self):
        return b""

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type: pa.DataType, serialized: bytes):
        return cls(interleaved=True, dims=CoordinateDimension.XY)


class LineStringType(BaseGeometryType):
    extension_name = "geoarrow.linestring"

    def __init__(
        self, *, interleaved: bool, dims: CoordinateDimension, large_list: bool = False
    ):
        self.coord_dimension = dims

        storage_type = linestring_storage_type(
            interleaved=interleaved, dims=dims, large_list=large_list
        )
        super().__init__(storage_type, self.extension_name)

    def __arrow_ext_serialize__(self):
        return b""

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type: pa.DataType, serialized: bytes):
        return cls(interleaved=True, dims=CoordinateDimension.XY)


class PolygonType(BaseGeometryType):
    extension_name = "geoarrow.polygon"

    def __init__(
        self, *, interleaved: bool, dims: CoordinateDimension, large_list: bool = False
    ):
        self.coord_dimension = dims

        storage_type = polygon_storage_type(
            interleaved=interleaved, dims=dims, large_list=large_list
        )
        super().__init__(storage_type, self.extension_name)

    def __arrow_ext_serialize__(self):
        return b""

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type: pa.DataType, serialized: bytes):
        return cls(interleaved=True, dims=CoordinateDimension.XY)


class MultiPointType(BaseGeometryType):
    extension_name = "geoarrow.multipoint"

    def __init__(
        self, *, interleaved: bool, dims: CoordinateDimension, large_list: bool = False
    ):
        self.coord_dimension = dims

        storage_type = multipoint_storage_type(
            interleaved=interleaved, dims=dims, large_list=large_list
        )
        super().__init__(storage_type, self.extension_name)

    def __arrow_ext_serialize__(self):
        return b""

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type: pa.DataType, serialized: bytes):
        return cls(interleaved=True, dims=CoordinateDimension.XY)


class MultiLineStringType(BaseGeometryType):
    extension_name = "geoarrow.multilinestring"

    def __init__(
        self, *, interleaved: bool, dims: CoordinateDimension, large_list: bool = False
    ):
        self.coord_dimension = dims

        storage_type = multilinestring_storage_type(
            interleaved=interleaved, dims=dims, large_list=large_list
        )
        super().__init__(storage_type, self.extension_name)

    def __arrow_ext_serialize__(self):
        return b""

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type: pa.DataType, serialized: bytes):
        return cls(interleaved=True, dims=CoordinateDimension.XY)


class MultiPolygonType(BaseGeometryType):
    extension_name = "geoarrow.multipolygon"

    def __init__(
        self, *, interleaved: bool, dims: CoordinateDimension, large_list: bool = False
    ):
        self.coord_dimension = dims

        storage_type = multipolygon_storage_type(
            interleaved=interleaved, dims=dims, large_list=large_list
        )
        super().__init__(storage_type, self.extension_name)

    def __arrow_ext_serialize__(self):
        return b""

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type: pa.DataType, serialized: bytes):
        return cls(interleaved=True, dims=CoordinateDimension.XY)


def construct_geometry_array(
    shapely_arr: NDArray[np.object_],
    include_z: Optional[bool] = None,
    *,
    crs_str: Optional[str] = None,
) -> Tuple[pa.Field, pa.Array]:
    # NOTE: this implementation returns a (field, array) pair so that it can set the
    # extension metadata on the field without instantiating extension types into the
    # global pyarrow registry
    geom_type, coords, offsets = shapely.to_ragged_array(
        shapely_arr, include_z=include_z
    )

    if coords.shape[-1] == 2:
        dims = CoordinateDimension.XY
    elif coords.shape[-1] == 3:
        dims = CoordinateDimension.XYZ
    else:
        raise ValueError(f"Unexpected coords dimensions: {coords.shape}")

    extension_metadata: Dict[str, str] = {}
    if crs_str is not None:
        extension_metadata["ARROW:extension:metadata"] = json.dumps({"crs": crs_str})

    if geom_type == GeometryType.POINT:
        parr = pa.FixedSizeListArray.from_arrays(coords.flatten(), len(dims))
        extension_metadata["ARROW:extension:name"] = "geoarrow.point"
        field = pa.field(
            "geometry",
            parr.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, parr

    elif geom_type == GeometryType.LINESTRING:
        assert len(offsets) == 1, "Expected one offsets array"
        (geom_offsets,) = offsets
        _parr = pa.FixedSizeListArray.from_arrays(coords.flatten(), len(dims))
        parr = pa.ListArray.from_arrays(pa.array(geom_offsets), _parr)
        extension_metadata["ARROW:extension:name"] = "geoarrow.linestring"
        field = pa.field(
            "geometry",
            parr.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, parr

    elif geom_type == GeometryType.POLYGON:
        assert len(offsets) == 2, "Expected two offsets arrays"
        ring_offsets, geom_offsets = offsets
        _parr = pa.FixedSizeListArray.from_arrays(coords.flatten(), len(dims))
        _parr1 = pa.ListArray.from_arrays(pa.array(ring_offsets), _parr)
        parr = pa.ListArray.from_arrays(pa.array(geom_offsets), _parr1)
        extension_metadata["ARROW:extension:name"] = "geoarrow.polygon"
        field = pa.field(
            "geometry",
            parr.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, parr

    elif geom_type == GeometryType.MULTIPOINT:
        assert len(offsets) == 1, "Expected one offsets array"
        (geom_offsets,) = offsets
        _parr = pa.FixedSizeListArray.from_arrays(coords.flatten(), len(dims))
        parr = pa.ListArray.from_arrays(pa.array(geom_offsets), _parr)
        extension_metadata["ARROW:extension:name"] = "geoarrow.multipoint"
        field = pa.field(
            "geometry",
            parr.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, parr

    elif geom_type == GeometryType.MULTILINESTRING:
        assert len(offsets) == 2, "Expected two offsets arrays"
        ring_offsets, geom_offsets = offsets
        _parr = pa.FixedSizeListArray.from_arrays(coords.flatten(), len(dims))
        _parr1 = pa.ListArray.from_arrays(pa.array(ring_offsets), _parr)
        parr = pa.ListArray.from_arrays(pa.array(geom_offsets), _parr1)
        extension_metadata["ARROW:extension:name"] = "geoarrow.multilinestring"
        field = pa.field(
            "geometry",
            parr.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, parr

    elif geom_type == GeometryType.MULTIPOLYGON:
        assert len(offsets) == 3, "Expected three offsets arrays"
        ring_offsets, polygon_offsets, geom_offsets = offsets
        _parr = pa.FixedSizeListArray.from_arrays(coords.flatten(), len(dims))
        _parr1 = pa.ListArray.from_arrays(pa.array(ring_offsets), _parr)
        _parr2 = pa.ListArray.from_arrays(pa.array(polygon_offsets), _parr1)
        parr = pa.ListArray.from_arrays(pa.array(geom_offsets), _parr2)
        extension_metadata["ARROW:extension:name"] = "geoarrow.multipolygon"
        field = pa.field(
            "geometry",
            parr.type,
            nullable=True,
            metadata=extension_metadata,
        )
        return field, parr

    else:
        raise ValueError(f"Unsupported type for geoarrow: {geom_type}")
